from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union
import yaml

from opa_tools.config import get_request_dir
from .transformation import flatten
from .utils import RequestIdentifierError


REQUEST_DIR = get_request_dir()
ALLOWED_KINDS = (
    "gsv",
    "opa",
)


def _normalize_request_name(name: Any):
    # Can raise type error. # TODO: Make a user error
    return f"{int(name):03d}"


def _validate_month(month: str):
    if not 1 <= int(month) <= 12:
        raise ValueError(f"Bad value for month: {month}.")


def _normalize_request_month(month: str | int | None, identifier: Any) -> str:
    now = datetime.now()
    if month is None:
        return now.strftime("%y%m")
    month = str(month)
    if not month.isdigit():
        raise ValueError(f"Value for month must be a number, but was '{month}'.")
    if len(month) == 2:
        _validate_month(month)
        return now.strftime("%y") + month
    elif len(month) == 4:
        _validate_month(month[2:4])
        return month
    raise RequestIdentifierError(identifier)


def _normalize_request_identifier(
    identifier: Any, month: Optional[Union[str, int]] = None, kind: str = "*"
) -> Tuple[str, str, str]:
    if isinstance(identifier, int):
        identifier = str(identifier)
    elif isinstance(identifier, str):
        pass
    else:
        raise TypeError(f"Bad type for request identifier: '{type(identifier)}'.")
    # Split identifier
    if "/" in identifier:
        month, name = identifier.split("/", 1)
    else:
        name = identifier
        month = _normalize_request_month(month, identifier)
    # Normalize triplet
    name = _normalize_request_name(name)
    month = _normalize_request_month(month, identifier)
    if kind not in ["*", *ALLOWED_KINDS]:
        raise ValueError(f"Wrong kind: '{kind}', must be in {ALLOWED_KINDS}.")
    return name, month, kind


def _get_single_request(name: str, month: str, kind: str) -> Dict:
    if kind not in ("opa", "gsv"):
        raise AssertionError(f"kind must be either 'opa' or 'gsv', but was '{kind}'.")
    request_path = REQUEST_DIR / month / name / f"{kind}.yml"
    with open(request_path, "r", encoding="utf-8") as rf:
        return yaml.safe_load(rf)


def get_request_identifiers(month: str = "*"):
    if month == "*":
        output = []
        for month in sorted(path.name for path in REQUEST_DIR.iterdir()):
            output.extend(get_request_identifiers(month=month))
        return output
    output = []
    for path in sorted((REQUEST_DIR / month).iterdir()):
        output.append(path.parent.name + "/" + path.name)
    return output


def get_request(
    identifier: Optional[Union[str, int]] = None,
    /,
    month: Optional[str] = None,
    kind: str = "*",
    *,
    flat=False,
    single_as_list=False,
) -> Union[List[Dict], Dict]:
    if identifier is None:
        if month is None:
            month = "*"
        output = []
        for identifier in get_request_identifiers(month=month):
            output.append(
                get_request(identifier, kind=kind, flat=flat, single_as_list=True)
            )
        return output
    name, month, kind = _normalize_request_identifier(identifier, month, kind)
    if kind == "*":
        output = [
            _get_single_request(name=name, month=month, kind="gsv"),
            _get_single_request(name=name, month=month, kind="opa"),
        ]
        if flat:
            output = flatten(*output)
        return output
    if single_as_list:
        return [_get_single_request(name=name, month=month, kind=kind)]
    return _get_single_request(name=name, month=month, kind=kind)
